observableAnswers(pos(subst), A, Epos) :-
    oracle(pos(subst), A, F), posubToPos(F,Epos).

observableClause(pos(subst), A, Epos, B, C) :-
    observableClause(subst, A, E, B, C), posubEqua((A:-B), E, F), posubToPos(F,Ebig), varsOf((A,B),Vars), posProj(Ebig, Vars, Epos).

domainBottom(pos(subst), X) :- posBottom(X).
    
domainJoin(pos(subst), F1, F2, F) :- posJoin(F1, F2, F).

domainMeet(pos(subst), F1, F2, F) :- posMeet(F1, F2, F).
    
domainProject(pos(subst), A, F, Fr) :-
    varsOf(A,Vars), posProj(F, Vars, Fr).

domainSingleton(pos(subst), F, Fs) :- posImplicant(F, Fs).

readObs(pos(subst),A,E) :- read(X), (X \== no -> X=(A,E); X=E),assert( (oracleMemory(Obs, A, no_more_ans)) ).

showAelement(pos(subst),A,or([E]),(A,E)):-!.
showAelement(pos(subst),A,E,(A,E)).

varsOf(T,Xs) :- varsof(T,Ys),removeReps([],_SeenOut,Ys,Xs).
varsof(X,[X]) :- var(X),!.
varsof(A,Vars) :- A =.. [_|L], varsof(L,[],Vars).
varsof([],L,L).
varsof([A|As],L,Vs) :- varsof(A,AL),append(AL,L,G),varsof(As,G,Vs).

posubEqua((A:-(B,C)),(EA:-(EB,EC)),and(F1,F2)) :-
    !,posubEqua((A:-B),(EA:-EB),F1),posubEqua(C,EC,F2).
posubEqua((A:-B),(EA:-EB),and(F1,F2)) :-
    !,posubEqua(A,EA,F1),posubEqua(B,EB,F2).
posubEqua(true,true,true):-!.
posubEqua(A, E, F):-
    functor(A,Fun,N),functor(E,Fun,N), posubEqua(N, A, E, F).

posubEqua(1, A, E, Fiff) :- arg(1,A,V), arg(1,E,T), posubEquaIff(V, T, Fiff).
posubEqua(N, A, E, and(Fiff,F2)) :- 
    N >1, arg(N,A,V), arg(N,E,T), posubEquaIff(V, T, Fiff), 
    N1 is N-1, posubEqua(N1, A, E, F2).

posubEquaIff(V, T, V) :- posubTerm(T, true),!.
posubEquaIff(V, T, iff(V,F)) :- posubTerm(T, F).

posubTerm(T,true) :- atomic(T),!.
posubTerm(T,T) :- var(T),!.
posubTerm(T, F) :- compound(T),functor(T,_Name,N), posubTerm(N, T, F).

posubTerm(1, T, F):- arg(1, T, A), posubTerm(A, F).
posubTerm(N, T, F) :- 
    N >1, arg(N, T, A), posubTerm(A, true), !, N1 is N-1, posubTerm(N1, T, F).
posubTerm(N, T, F) :- 
    N >1, N1 is N-1, posubTerm(N1, T, true), !, arg(N, T, A), posubTerm(A, F).
posubTerm(N, T, and(F1,F2)) :- 
    N >1, arg(N, T, A), posubTerm(A, F1), N1 is N-1, posubTerm(N1, T, F2).

posubToPos(X,F):- var(X),!,posVar(X,F).
posubToPos(true,X) :- posTop(X).
posubToPos(false,X) :- posBottom(X).
posubToPos(and(F1,F2),V) :-
    posubToPos(F1,V1), posubToPos(F2,V2), posMeet(V1,V2,V).
posubToPos(or(F1,F2),V) :-
    posubToPos(F1,V1), posubToPos(F2,V2), posJoin(V1,V2,V).
posubToPos(iff(Var,F),R) :- posubToPos(F,V), posIff(Var,V,R).

/* POS domain */

posTop(or([and([])])).

posBottom(or([])).

posVar(X,or([and([X])])).

/* posIff */

posIff(Var,or([and(T)]),or([ and([Var|T]) |Cs])) :- onlyIf(T,Var,or(Cs)).

onlyIf([],_Var,or([])).
onlyIf([Hneg|T],Var,or([ and([-Var,H]) |Cs])) :-
    nonvar(Hneg), Hneg = -H, !, onlyIf(T,Var,or(Cs)).
onlyIf([H|T],Var,or([ and([-Var,-H]) |Cs])) :- onlyIf(T,Var,or(Cs)).

/* posJoin */

posJoin(or(L1),or(L2),C) :- append(L1,L2,L), simplify2(or(L),C1), simplify3(C1,C2),simplify2(C2,C).

simplify2(C,CS) :-
    powersOfTwo(P), compNums(C,P,Nums,[],_), selectEntailed([],1,Nums,Ent), sort(Ent,OrdEnt), removeEntailed(1,C,OrdEnt,CS).

powersOfTwo([1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,
32768,65536,131072,262144,524288,1048576,2097152,4194304,8388608,
16777216,33554432,67108864,134217728]).

compNums(or([]),_Table,[],_Seen, []).
compNums(or([H|T]),Table,[NH|NT],Seen,Lout) :-
    compNums(H,NH,Table,TableOut,Seen,SeenOut,Lin,Lout), 
    compNums(or(T),TableOut,NT,SeenOut,Lin).

compNums(and([]),0,Table,Table,Seen,Seen,Lin,Lin).
compNums(and([X]),Num,Table,Table,Seen,Seen,Lin,Lin) :- 
    funnyMember((X,Num),Seen),!.
compNums(and([X]),Num,[Num|TT],TT,Seen,[(X,Num)|Seen],Lin,Lout) :- 
    !, update(X,Num,Seen,Lin,Lout).
compNums(and([X|T]),Num,Table,TableOut,Seen,SeenOut,Lin,Lout) :-
    funnyMember((X,Num1),Seen), !, compNums(and(T),Num2,Table,TableOut,Seen,SeenOut,Lin,Lout), Num is Num1+Num2.
compNums(and([X|T]),Num,[Num1|TT],TableOut,Seen,SeenOut,Lin,Lout) :-
    update(X,Num1,Seen,Lin,L1), compNums(and(T),Num2,TT,TableOut,[(X,Num1)|Seen],SeenOut,L1,Lout), Num is Num1+Num2.

update(X, Num1, Seen, L, [(X,Num1,Num2)|L]) :- var(X), funnyMember((-X,Num2),Seen),!.
update(-X, Num1, Seen, L, [(X,Num1,Num2)|L]) :- var(X), funnyMember((X,Num2),Seen),!.
update(_, _, _, L, L).

funnyMember((Var,Num),[(Var1,Num)|_T]) :- Var==Var1,!.
funnyMember(X,[_H|T]) :- funnyMember(X,T).

selectEntailed(_Seen,_K,[],[]).
selectEntailed(Seen,K,[H|T],[K|T1]) :- 
    isEntailed(H,Seen), !, K1 is K+1, selectEntailed(Seen,K1,T,T1).
selectEntailed(Seen,K,[H|T],Result) :-
    entailed(H,Seen,E), K1 is K+1, append(E,T1,Result), selectEntailed([(H,K)|Seen],K1,T,T1).

isEntailed(N,[(M,_Pos)|_T]) :- entails(M,N), !.
isEntailed(N,[_H|T]) :- isEntailed(N,T).

entailed(_N,[],[]).
entailed(N,[(M,Pos)|T],[Pos|T1]) :- entails(N,M), !, entailed(N,T,T1).
entailed(N,[_H|T],T1) :- entailed(N,T,T1).

entails(N,NEntailed) :- Tmp is N /\ NEntailed,Tmp=N.

removeEntailed(_K,or([]),[],or([])).
removeEntailed(K,or([_H|T]),[K|TE],or(T1)) :-
    !, K1 is K+1, removeEntailed(K1,or(T),TE,or(T1)).
removeEntailed(K,or([H|T]),Entailed,or([H|T1])) :-
    K1 is K+1, removeEntailed(K1,or(T),Entailed,or(T1)).

simplify3(or(L),or(LS)) :- powersOfTwo(P), compNums(or(L),P,Nums,[],Sus), joinNumsC(Sus,Nums,L,_,LS).

joinNumsC([],Ns,As,Ns,As).
joinNumsC([_|_],[],[],[],[]).
joinNumsC([(X,P1,P2)|Ds],[N|Ns],[A|As],N3s,A3s) :- joinNumsC(X,P1,P2,N,Ns,A,As,N2,N1s,A2,A1s), joinNumsC([(X,P1,P2)|Ds],N1s,A1s,N2s,A2s), joinNumsC(Ds,[N2|N2s],[A2|A2s],N3s,A3s).

joinNumsC(_,_,_,N,[],A,[],N,[],A,[]).
joinNumsC(X,P1,P2,N1,[N2|Ns],and(L1),[_A2|As],N3,Ns,and(L2),As) :- canjoin(N1,N2,P1,P2), !, N3 is (N1 /\ \(P1)) /\ \(P2), softDelete(X, L1, L2).
joinNumsC(X,P1,P2,N1,[N|Ns],A1,[A|As],N2,[N|N2s],A2,[A|A2s]) :- joinNumsC(X,P1,P2,N1,Ns,A1,As,N2,N2s,A2,A2s).

canjoin(N1,N2,P1,P2) :- P is P1+P2, N is (N1 /\ \(N2)) \/ (N2 /\ \(N1)), P=N, (X is N1 /\ P, X=P1, Y is N2 /\ P, Y=P2 ; X is N1 /\ P, X=P2, Y is N2 /\ P, Y=P1).

softDelete(X, [Y|Ys], Ys) :- X==Y, !.
softDelete(X, [Y|Ys], [Y|Zs]) :- softDelete(X,Ys,Zs).

/* posMeet */

posMeet(or([]),_,or([])).
posMeet(or([H|T]),L,Result) :-
    product(H,L,or(H1)), append(H1,T1,Res), posMeet(or(T),L,or(T1)), simplify1(or(Res),Resu), simplify2(Resu,Resul), simplify3(Resul,Result).

product(and(_L),or([]),or([])).
product(and(L),or([and(H)|T]),or([and(H1)|T1])) :-
    append(L,H,H1), product(and(L),or(T),or(T1)).

simplify1(or([]),or([])).
simplify1(or([H|T]),or(Result)):-
    simplify1Aux([],H,H1,Flag), (Flag=yes -> Result=T1 ; Result=[H1|T1]), simplify1(or(T),or(T1)).

simplify1Aux(_Seen,and([]),and([]),no).
simplify1Aux(Seen,and([H|T]),and(T1),Flag):-
    memberVar(H,Seen,How), (How=pos -> simplify1Aux([H|Seen],and(T),and(T1),Flag) ; (How=neg -> Flag=yes ; (simplify1Aux([H|Seen],and(T),and(T2),Flag),T1=[H|T2]))).

memberVar(_Var,[],no).
memberVar(Var,[H|_T],pos) :- Var==H,!.
memberVar(-Var,[H|_T],neg) :- Var==H,!.
memberVar(Var,[-H|_T],neg) :- Var==H,!.
memberVar(Var,[_H|T],How) :- memberVar(Var,T,How).

/* posProj */

posProj(Const,Vars,Result):-
    constVars(Const,LV), diff(LV,Vars,Diff), posRestrict(Const,Diff,Res), simplify3(Res,Result).

constVars(or(List),Xs):-
    constVarsWRep(List,Ys),removeReps([],_SeenOut,Ys,Xs).

constVarsWRep([],[]).
constVarsWRep([and(C)|Cs],Rs) :- constVarsWRep(Cs,Ds),constVarsWRep(C,Ds,Rs).

constVarsWRep([],Ds,Ds).
constVarsWRep([Hneg|Cs],Ds,[H|Rs]) :-
    nonvar(Hneg), Hneg = -H, !, constVarsWRep(Cs,Ds,Rs).
constVarsWRep([H|Cs],Ds,[H|Rs]) :- constVarsWRep(Cs,Ds,Rs).

removeReps(Seen,Seen,[],[]).
removeReps(SeenIn,SeenOut,[H|T],T1) :-
    softMember(H,SeenIn),!, removeReps(SeenIn,SeenOut,T,T1).
removeReps(SeenIn,SeenOut,[H|T],[H|T1]) :- removeReps([H|SeenIn],SeenOut,T,T1).

softMember(X,[Y|_L]) :- X==Y,!.
softMember(X,[_Y|L]) :- softMember(X,L).

diff([],_L,[]).
diff([H|T],L,T1) :- softMember(H,L), !, diff(T,L,T1).
diff([H|T],L,[H|T1]) :- diff(T,L,T1).

posRestrict(C,[],C).
posRestrict(C,[H|T],CR):-
    schroder(C,H,CS1), simplify1(CS1,CS2), simplify2(CS2,CS), posRestrict(CS,T,CR).

schroder(or([]),_V,or([])).
schroder(or([ and(A) | O]),V,or([ and(AwoV) | OwoV])) :-
    kill(V,A,AwoV), schroder(or(O),V,or(OwoV)).

kill(_V,[],[]).
kill(V,[X|Xs],Xs) :- X == V.
kill(V,[X|Xs],Xs) :- X == -V.
kill(V,[X|Xs],[X|Ys]) :- X \== V, X \== -V, kill(V,Xs,Ys).

/* posImplicant */

posImplicant(or([F]),or(Fs)) :- var(F), member(F,Fs).
posImplicant(or([F]),or(Fs)) :-
    nonvar(F), powersOfTwo(P), compNums(or([F|Fs]),P,[N|Ns],[],_), memberEntails(N,Ns).

memberEntails(V,[N|_Ns]) :- entails(N,V),!.
memberEntails(V,[_N|Ns]) :- memberEntails(V,Ns).
